"""Manages configuration for argparse"""
import argparse


def add_target_argument(parser: argparse.ArgumentParser) -> None:
    """Adds --target arg to a subparser"""
    parser.add_argument(
        "-t",
        "--target",
        action="append",
        help=(
            "Specify one or more studies to perform actions against. "
            "Default is to use all studies."
        ),
    )


def add_table_builder_argument(parser: argparse.ArgumentParser) -> None:
    """Adds --table_builder arg to a subparser"""
    parser.add_argument(
        "--table-builder",
        help=(argparse.SUPPRESS),
    )


def add_study_dir_argument(parser: argparse.ArgumentParser) -> None:
    """Adds --study_dir arg to a subparser"""
    parser.add_argument(
        "-s",
        "--study-dir",
        action="append",
        help=(
            "Optionally add one or more directories to look for study definitions in. "
            "Default is in project directory and CUMULUS_LIBRARY_STUDY_DIR, if present, "
            "followed by any supplied paths. Target, and all its subdirectories, "
            "are checked for manifests. Overriding studies with the same namespace "
            "supersede earlier ones."
        ),
    )


def add_data_path_argument(parser: argparse.ArgumentParser) -> None:
    """Adds path arg to a subparser"""
    parser.add_argument(
        "data_path",
        default="./",
        nargs="?",
        help=(
            "The path to use for Athena counts data. "
            "Can be povided via CUMULUS_LIBRARY_DATA_PATH environment variable."
        ),
    )


def add_verbose_argument(parser: argparse.ArgumentParser) -> None:
    """Adds --verbose arg to a subparser"""
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="Prints detailed SQL query info",
    )


def add_aws_config(parser: argparse.ArgumentParser) -> None:
    """Adds arguments related to aws credentials to a subparser"""
    aws = parser.add_argument_group("AWS config")
    aws.add_argument("--profile", help="AWS profile", default="default")
    aws.add_argument(
        "--database",
        # internally, we use PyAthena's terminology for this but the UX term is "database"
        dest="schema_name",
        help="Cumulus Athena database name",
    )
    aws.add_argument(
        "--workgroup",
        default="cumulus",
        help="Cumulus Athena workgroup (default: cumulus)",
    )
    aws.add_argument(
        "--region",
        help="AWS region data of Athena instance (default: us-east-1)",
        default="us-east-1",
    )


def get_parser() -> argparse.ArgumentParser:
    """Provides parser for handling CLI arguments"""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Generates study tables and views from post-Cumulus ETL data.

cumulus-library will attempt to create a connection to AWS Athena. The
following order of preference is used to select credentials:
  - explicit command line arguments
  - cumulus environment variables (see docs for more info)
  - Normal boto profile order (AWS env vars, ~/.aws/credentials, ~/.aws/config)""",
        epilog="See 'cumulus-library -h [action]' for usage of a specific action",
    )

    parser.add_argument(
        "--version", action="store_true", help="Display cumulus-library version number"
    )

    actions = parser.add_subparsers(
        title="actions",
        help="Available library actions",
        dest="action",
    )

    create = actions.add_parser(
        "create", help="Create a study instance from a template"
    )
    create.add_argument(
        "create_dir",
        default="./",
        nargs="?",
        help=(
            "The the directory the study will be created in. Default is "
            "the current directory."
        ),
    )

    clean = actions.add_parser(
        "clean", help="Removes tables & views beginning with '[target]__' from Athena"
    )
    add_target_argument(clean)
    add_aws_config(clean)

    build = actions.add_parser(
        "build",
        help="Removes and recreates Athena tables & views for specified studies",
    )
    add_target_argument(build)
    add_table_builder_argument(build)
    add_study_dir_argument(build)
    add_verbose_argument(build)
    add_aws_config(build)

    export = actions.add_parser(
        "export", help="Generates files on disk from Athena views"
    )
    add_target_argument(export)
    add_study_dir_argument(export)
    add_data_path_argument(export)
    add_verbose_argument(export)
    add_aws_config(export)

    upload = actions.add_parser(
        "upload", help="Bulk uploads data to Cumulus aggregator"
    )
    add_target_argument(upload)
    add_data_path_argument(upload)
    upload.add_argument(
        "--user", help="Cumulus user. Default is value of CUMULUS_AGGREGATOR_USER"
    )
    upload.add_argument(
        "--id", help="Site ID. Default is value of CUMULUS_AGGREGATOR_ID"
    )
    upload.add_argument(
        "--url",
        help=(
            "Upload URL. Default is value of CUMULUS_AGGREGATOR_URL if present, "
            "or smart cumulus instance"
        ),
        default="https://aggregator.smartcumulus.org/upload/",
    )
    upload.add_argument(
        "--preview",
        default=False,
        action="store_true",
        help="Run pre-fetch and prepare upload, but log output instead of sending.",
    )

    return parser



""" Contains classes for loading study data based on manifest.toml files """
import inspect
import importlib.util
import sys

from pathlib import Path, PosixPath
from typing import List, Optional

import toml

from rich.progress import Progress, TaskID, track

from cumulus_library.base_table_builder import BaseTableBuilder
from cumulus_library.errors import StudyManifestParsingError
from cumulus_library.helper import (
    query_console_output,
    load_text,
    parse_sql,
    get_progress_bar,
)
from cumulus_library.template_sql.templates import (
    get_show_tables,
    get_show_views,
    get_drop_view_table,
)

StrList = List[str]

RESERVED_TABLE_KEYWORDS = ["etl", "nlp", "lib"]


class StudyManifestParser:
    """Handles loading of study data from manifest files.

    The goal of this class is to make it so that a researcher can contribute a study
    definition without touching the main python infrastructure. It provides
    mechanisms for IDing studies/files of interest, and for executing queries, but
    specifically it should never be in charge of instantiation a cursor itself -
    this will help to future proof against other database implementations in the
    future, assuming those DBs have a PEP-249 cursor available (and this is why we
    are hinting generic objects for cursors).

    """

    _study_path = None
    _study_config = {}

    def __init__(self, study_path: Optional[Path] = None):
        """Instantiates a StudyManifestParser.

        :param study_path: A pathlib Path object, optional
        """
        if study_path is not None:
            self.load_study_manifest(study_path)

    def __repr__(self):
        return str(self._study_config)

    ### toml parsing helper functions
    def load_study_manifest(self, study_path: Path) -> None:
        """Reads in a config object from a directory containing a manifest.toml

        :param study_path: A pathlib.Path object pointing to a study directory
        :raises StudyManifestParsingError: the manifest.toml is malformed or missing.
        """
        try:
            with open(f"{study_path}/manifest.toml", encoding="UTF-8") as file:
                config = toml.load(file)
                if not config.get("study_prefix") or not isinstance(
                    config["study_prefix"], str
                ):
                    raise StudyManifestParsingError(
                        f"Invalid prefix in manifest at {study_path}"
                    )
                self._study_config = config
            self._study_path = study_path
        except FileNotFoundError:
            raise StudyManifestParsingError(  # pylint: disable=raise-missing-from
                f"Missing or invalid manifest found at {study_path}"
            )

    def get_study_prefix(self) -> Optional[str]:
        """Reads the name of a study prefix from the in-memory study config

        :returns: A string of the prefix in the manifest, or None if not found
        """
        return self._study_config.get("study_prefix")

    def get_sql_file_list(self) -> Optional[StrList]:
        """Reads the contents of the sql_config array from the manifest

        :returns: An array of sql files from the manifest, or None if not found.
        """
        sql_config = self._study_config.get("sql_config", {})
        return sql_config.get("file_names", [])

    def get_table_builder_file_list(self) -> Optional[StrList]:
        """Reads the contents of the python_config array from the manifest

        :returns: An array of sql files from the manifest, or None if not found.
        """
        sql_config = self._study_config.get("table_builder_config", {})
        return sql_config.get("file_names", [])

    def get_export_table_list(self) -> Optional[StrList]:
        """Reads the contents of the export_list array from the manifest

        :returns: An array of tables to export from the manifest, or None if not found.
        """
        export_config = self._study_config.get("export_config", {})
        export_table_list = export_config.get("export_list", [])
        for table in export_table_list:
            if not table.startswith(f"{self.get_study_prefix()}__"):
                raise StudyManifestParsingError(
                    f"{table} in export list does not start with prefix "
                    f"{self.get_study_prefix()}__ - check your manifest file."
                )
        return export_table_list

    def reset_export_dir(self) -> None:
        """
        Removes exports associated with this study from the ../data_export directory.
        """
        project_path = Path(__file__).resolve().parents[1]
        path = Path(f"{str(project_path)}/data_export/{self.get_study_prefix()}/")
        if path.exists():
            for file in path.glob("*"):
                file.unlink()

    # SQL related functions
    def clean_study(
        self,
        cursor: object,
        schema_name: str,
        verbose: bool = False,
        prefix: str = None,
    ) -> List:
        """Removes tables beginning with the study prefix from the database schema

        :param cursor: A PEP-249 compatible cursor object
        :param schema_name: The name of the schema containing the study tables
        :verbose: toggle from progress bar to query output, optional
        :returns: list of dropped tables (for unit testing only)

        TODO: If we need to support additional databases, we may need to investigate
        additional ways to get a list of table prefixes
        """
        if not schema_name:
            raise ValueError("No database provided")
        if not prefix:
            prefix = self.get_study_prefix()
        view_sql = get_show_views(schema_name, prefix)
        table_sql = get_show_tables(schema_name, prefix)
        view_table_list = []
        for query_and_type in [[view_sql, "VIEW"], [table_sql, "TABLE"]]:
            cursor.execute(query_and_type[0])
            for db_row_tuple in cursor:
                # this check handles athena reporting views as also being tables,
                # so we don't waste time dropping things that don't exist
                if query_and_type[1] == "TABLE":
                    if not any(
                        db_row_tuple[0] in query_and_type
                        for query_and_type in view_table_list
                    ):
                        view_table_list.append([db_row_tuple[0], query_and_type[1]])
                else:
                    view_table_list.append([db_row_tuple[0], query_and_type[1]])
        if not view_table_list:
            return view_table_list

        # We'll do a pass to see if any of these tables were created outside of a
        # study builder, and remove them from the list.
        for view_table in view_table_list.copy():
            if any(
                view_table[0].startswith(f"{self.get_study_prefix()}__{word}_")
                for word in RESERVED_TABLE_KEYWORDS
            ):
                view_table_list.remove(view_table)

        # We want to only show a progress bar if we are :not: printing SQL lines
        with get_progress_bar(disable=verbose) as progress:
            task = progress.add_task(
                f"Removing {self.get_study_prefix()} study artifacts...",
                total=len(view_table_list),
                visible=not verbose,
            )
            self._execute_drop_queries(cursor, verbose, view_table_list, progress, task)
        return view_table_list

    def _execute_drop_queries(
        self,
        cursor: object,
        verbose: bool,
        view_table_list: List,
        progress: Progress,
        task: TaskID,
    ) -> None:
        """Handler for executing drop view/table queries and displaying console output.

        :param cursor: A PEP-249 compatible cursor object
        :param verbose: toggle from progress bar to query output
        :param view_table_list: a list of views and tables beginning with the study prefix
        :param progress: a rich progress bar renderer
        :param task: a TaskID for a given progress bar
        """
        for view_table in view_table_list:
            drop_view_table = get_drop_view_table(
                name=view_table[0], view_or_table=view_table[1]
            )
            cursor.execute(drop_view_table)
            query_console_output(verbose, drop_view_table, progress, task)

    def _load_and_execute_builder(
        self, filename, cursor, schema, verbose, drop_table=False
    ) -> None:
        """Loads a table builder from a file.

        Since we have to support arbitrary user-defined python files here, we
        jump through some importlib hoops to import the module directly from
        a source file defined in the manifest.

        As with eating an ortolan, you may wish to cover your head with a cloth.
        Per an article on the subject: "Tradition dictates that this is to shield
        – from God’s eyes – the shame of such a decadent and disgraceful act."

        """
        spec = importlib.util.spec_from_file_location(
            "table_builder", f"{self._study_path}/{filename}"
        )
        table_builder_module = importlib.util.module_from_spec(spec)
        sys.modules["table_builder"] = table_builder_module
        spec.loader.exec_module(table_builder_module)

        # We're going to find all subclasses of BaseTableBuild in this file.
        # Since BaseTableBuilder itself is a valid subclass of BaseTableBuilder,
        # we'll detect and skip it. If we don't find exactly one subclass,
        # we'll punt.
        table_builder_subclasses = []
        for _, cls_obj in inspect.getmembers(table_builder_module, inspect.isclass):
            if issubclass(cls_obj, BaseTableBuilder) and cls_obj != BaseTableBuilder:
                table_builder_subclasses.append(cls_obj)
        if len(table_builder_subclasses) != 1:
            raise StudyManifestParsingError(
                f"Error loading {self._study_path}{filename}\n"
                "Custom builders must extend the BaseTableBuilder "
                "class exactly once per module."
            )

        # We'll get the subclass, initialize it, run the executor code, and then
        # remove it so it doesn't interfere with the next python module to
        # execute, since the subclass would otherwise hang around.
        table_builder_class = table_builder_subclasses[0]
        table_builder = table_builder_class()
        table_builder.execute_queries(cursor, schema, verbose, drop_table)

        # After runnning the executor code, we'll remove
        # remove it so it doesn't interfere with the next python module to
        # execute, since the subclass would otherwise hang around.
        del sys.modules[table_builder_module.__name__]
        del table_builder_module

    def run_table_builder(
        self, cursor: object, schema: str, verbose: bool = False
    ) -> None:
        """Loads modules from a manifest and executes code via BaseTableBuilder

        :param cursor: A PEP-249 compatible cursor object
        :param schema: The name of the schema to write tables to
        :param verbose: toggle from progress bar to query output
        """
        for file in self.get_table_builder_file_list():
            self._load_and_execute_builder(file, cursor, schema, verbose)

    def run_single_table_builder(
        self, cursor: object, schema: str, name: str, verbose: bool = False
    ):
        """targets a single table builder to run"""
        if not name.endswith(".py"):
            name = f"{name}.py"
        self._load_and_execute_builder(name, cursor, schema, verbose, drop_table=True)

    def build_study(self, cursor: object, verbose: bool = False) -> List:
        """Creates tables in the schema by iterating through the sql_config.file_names

        :param cursor: A PEP-249 compatible cursor object
        :param schema: The name of the schema to write tables to
        :param verbose: toggle from progress bar to query output, optional
        :returns: loaded queries (for unit testing only)
        """
        queries = []
        for file in self.get_sql_file_list():
            for query in parse_sql(load_text(f"{self._study_path}/{file}")):
                queries.append([query, file])
        if len(queries) == 0:
            return []
        # We want to only show a progress bar if we are :not: printing SQL lines
        with get_progress_bar(disable=verbose) as progress:
            task = progress.add_task(
                f"Creating {self.get_study_prefix()} study in db...",
                total=len(queries),
                visible=not verbose,
            )
            self._execute_build_queries(
                cursor,
                verbose,
                queries,
                progress,
                task,
            )
        return queries

    def _query_error(self, query_and_filename: List, exit_message: str) -> None:
        print(
            f"An error occured executing the following query in {query_and_filename[1]}:",
            file=sys.stderr,
        )
        print("--------", file=sys.stderr)
        print(query_and_filename[0], file=sys.stderr)
        print("--------", file=sys.stderr)
        sys.exit(exit_message)

    def _execute_build_queries(
        self,
        cursor: object,
        verbose: bool,
        queries: list,
        progress: Progress,
        task: TaskID,
    ) -> None:
        """Handler for executing create table queries and displaying console output.

        :param cursor: A PEP-249 compatible cursor object
        :param verbose: toggle from progress bar to query output
        :param queries: a list of queries read from files in sql_config.file_names
        :param progress: a rich progress bar renderer
        :param task: a TaskID for a given progress bar
        """
        for query in queries:
            create_line = query[0].split("\n")[0]
            if f" {self.get_study_prefix()}__" not in create_line:
                self._query_error(
                    query,
                    "This query does not contain the study prefix. All tables should "
                    f"start with a string like `{self.get_study_prefix()}__`, and it "
                    "should be in the first line of the query.",
                )
            if any(
                f" {self.get_study_prefix()}__{word}_" in create_line
                for word in RESERVED_TABLE_KEYWORDS
            ):
                self._query_error(
                    query,
                    "This query contains a table name which contains a reserved word "
                    "immediately after the study prefix. Please rename this table so "
                    "that is does not begin with one of these special words "
                    "immediately after the double undescore.\n"
                    f"Reserved words: {str(RESERVED_TABLE_KEYWORDS)}",
                )
            if create_line.count("__") > 1:
                self._query_error(
                    query,
                    "This query contains a table name with more than one '__' in it. "
                    "Double underscores are reserved for special use cases. Please "
                    "rename this table so the only double undercore is after the "
                    f"study prefix, e.g. `{self.get_study_prefix()}__`",
                )
            if f"{self.get_study_prefix()}__" not in query[0].split("\n")[0]:
                self._query_error(
                    query,
                    "This query does not contain the study prefix. All tables should "
                    "start with a string like `study_prefix__`.",
                )
            try:
                cursor.execute(query[0])
                query_console_output(verbose, query[0], progress, task)
            except Exception as e:  # pylint: disable=broad-exception-caught
                self._query_error(
                    query,
                    "You can debug issues with this query using `sqlfluff lint`, "
                    "or by executing the query directly against the database.\n"
                    f"Error: {e}",
                )

    # Database exporting functions

    def export_study(self, cursor: object, data_path: PosixPath) -> List:
        """Exports csvs/parquet extracts of tables listed in export_list

        :param cursor: A PEP-249 compatible cursor object
        :returns: list of executed queries (for unit testing only)

        TODO: If we need to support additional databases, we may need to investigate
        additional ways to convert the query to a pandas object
        """
        self.reset_export_dir()
        queries = []
        for table in track(
            self.get_export_table_list(),
            description=f"Exporting {self.get_study_prefix()} counts...",
        ):
            query = f"select * from {table}"
            dataframe = cursor.execute(query).as_pandas()
            path = Path(f"{str(data_path)}/{self.get_study_prefix()}/")
            path.mkdir(parents=True, exist_ok=True)
            dataframe.to_csv(f"{path}/{table}.csv", index=False)
            dataframe.to_parquet(f"{path}/{table}.parquet", index=False)
            queries.append(query)
        return queries



#!/usr/bin/env python3
"""Utility for building/retrieving data views in AWS Athena"""

import json
import os
import sys
import sysconfig

from pathlib import Path, PosixPath
from typing import Dict, List, Optional

import pyathena

from pyathena.pandas.cursor import PandasCursor
from rich.console import Console
from rich.table import Table

from cumulus_library import __version__
from cumulus_library.cli_parser import get_parser
from cumulus_library.study_parser import StudyManifestParser
from cumulus_library.upload import upload_files


# ** Don't delete! **
# This class isn't used in the rest of the code,
# but it is used manually as a quick & dirty alternative to the CLI.
class CumulusEnv:  # pylint: disable=too-few-public-methods
    """
    Wrapper for Cumulus Environment vars.
    Simplifies connections to StudyBuilder without requiring CLI parsing.
    """

    def __init__(self):
        self.region = os.environ.get("CUMULUS_LIBRARY_REGION", "us-east-1")
        self.workgroup = os.environ.get("CUMULUS_LIBRARY_WORKGROUP", "cumulus")
        self.profile = os.environ.get("CUMULUS_LIBRARY_PROFILE")
        self.schema_name = os.environ.get("CUMULUS_LIBRARY_DATABASE")

    def get_study_builder(self):
        """Convenience method for getting athena args from environment"""
        return StudyBuilder(self.region, self.workgroup, self.profile, self.schema_name)


class StudyBuilder:
    """Class for managing Athena cursors and executing Cumulus queries"""

    verbose = False
    schema_name = None

    def __init__(  # pylint: disable=too-many-arguments
        self, region: str, workgroup: str, profile: str, schema: str
    ):
        connect_kwargs = {}
        for aws_env_name in [
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
            "AWS_SESSION_TOKEN",
        ]:
            if aws_env_val := os.environ.get(aws_env_name):
                connect_kwargs[aws_env_name.lower()] = aws_env_val
        # the profile may not be required, provided the above three AWS env vars
        # are set. If both are present, the env vars take precedence
        if profile is not None:
            connect_kwargs["profile_name"] = profile
        self.cursor = pyathena.connect(
            region_name=region,
            work_group=workgroup,
            schema_name=schema,
            **connect_kwargs,
        ).cursor()
        self.pandas_cursor = pyathena.connect(
            region_name=region,
            work_group=workgroup,
            schema_name=schema,
            cursor_class=PandasCursor,
            **connect_kwargs,
        ).cursor()
        self.schema_name = schema

    def reset_data_path(self, study: PosixPath) -> None:
        """
        Removes existing exports from a study's local data dir
        """
        project_path = Path(__file__).resolve().parents[1]
        path = Path(f"{str(project_path)}/data_export/{study}/")
        if path.exists():
            for file in path.glob("*"):
                file.unlink()

    ### Creating studies

    def clean_study(self, targets: List[str]) -> None:
        """Removes study table/views from Athena.

        While this is usually not required, since it it done as part of a build,
        this can be useful for cleaning up tables if a study prefix is changed
        for some reason.

        :param target: The study prefix to use for IDing tables to remove"""
        if targets is None or targets == ["all"]:
            sys.exit(
                "Explicit targets for cleaning not provided. "
                "Provide one or more explicit study prefixes to remove."
            )
        for study in targets:
            StudyManifestParser.clean_study(
                self.cursor, self.schema_name, self.verbose, prefix=f"{study}__"
            )

    def clean_and_build_study(self, target: PosixPath) -> None:
        """Recreates study views/tables

        :param target: A PosixPath to the study directory
        """
        studyparser = StudyManifestParser(target)
        studyparser.clean_study(self.cursor, self.schema_name, self.verbose)
        studyparser.run_table_builder(self.cursor, self.schema_name, self.verbose)
        studyparser.build_study(self.cursor, self.verbose)

    def run_single_table_builder(
        self, target: PosixPath, table_builder_name: str
    ) -> None:
        """Runs a single table builder

        :param target: A PosixPath to the study directory
        """
        studyparser = StudyManifestParser(target)
        studyparser.run_single_table_builder(
            self.cursor, self.schema_name, table_builder_name, self.verbose
        )

    def clean_and_build_all(self, study_dict: Dict) -> None:
        """Builds views for all studies.

        NOTE: By design, this method will always exclude the `template` study dir,
        since 99% of the time you don't need a live copy in the database.

        :param study_dict: A dict of PosixPaths
        """
        study_dict = dict(study_dict)
        study_dict.pop("template")
        for precursor_study in ["vocab", "core"]:
            self.clean_and_build_study(study_dict[precursor_study])
            study_dict.pop(precursor_study)
        for key in study_dict:
            self.clean_and_build_study(study_dict[key])

    ### Data exporters
    def export_study(self, target: PosixPath, data_path: PosixPath) -> None:
        """Exports aggregates defined in a manifest

        :param target: A PosixPath to the study directory
        """
        if data_path is None:
            sys.exit("Missing destination - please provide a path argument.")
        studyparser = StudyManifestParser(target)
        studyparser.export_study(self.pandas_cursor, data_path)

    def export_all(self, study_dict: Dict, data_path: PosixPath):
        """Exports all defined count tables to disk"""
        for key in study_dict.keys():
            self.export_study(study_dict[key], data_path)


def get_abs_posix_path(path: str) -> PosixPath:
    """Convenience method for handling abs vs rel paths"""
    if path[0] == "/":
        return Path(path)
    else:
        return Path(Path.cwd(), path)


def create_template(path: str) -> None:
    """Creates a manifest in target dir if one doesn't exist"""
    abs_path = get_abs_posix_path(path)
    manifest_path = Path(abs_path, "manifest.toml")
    if manifest_path.exists():
        sys.exit(f"A manifest.toml already exists at {abs_path}, skipping creation")
    abs_path.mkdir(parents=True, exist_ok=True)

    copy_lists = [
        ["studies/template/manifest.toml", "manifest.toml"],
        [".sqlfluff", ".sqlfluff"],
    ]
    for source, dest in copy_lists:
        source_path = Path(Path(__file__).resolve().parents[0], source)
        dest_path = Path(abs_path, dest)
        dest_path.write_bytes(source_path.read_bytes())


def get_study_dict(alt_dir_paths: List) -> Optional[Dict[str, PosixPath]]:
    """Gets valid study targets from ./studies/, and any pip installed studies

    :returns: A list of pathlib.PosixPath objects
    """
    manifest_studies = {}
    cli_path = Path(__file__).resolve().parents[0]

    # first, we'll get any installed public studies
    with open(
        Path(cli_path, "./module_allowlist.json"), "r", encoding="utf-8"
    ) as study_allowlist_json:
        study_allowlist = json.load(study_allowlist_json)["allowlist"]
    site_packages_dir = sysconfig.get_path("purelib")
    for study, subdir in study_allowlist.items():
        study_path = Path(site_packages_dir, subdir)
        if study_path.exists():
            manifest_studies[study] = study_path

    # then we'll get all studies inside the project directory, followed by
    # any user supplied paths last. These take precedence.
    paths = [Path(cli_path, "studies")]
    if alt_dir_paths is not None:
        paths = paths + alt_dir_paths
    for parent_path in paths:
        for child_path in parent_path.iterdir():
            if child_path.is_dir():
                manifest_studies[child_path.name] = child_path
            elif child_path.name == "manifest.toml":
                manifest_studies[parent_path.name] = parent_path
    return manifest_studies


def run_cli(args: Dict):
    """Controls which library tasks are run based on CLI arguments"""
    if args["action"] == "create":
        create_template(args["create_dir"])

    elif args["action"] == "upload":
        upload_files(args)

    # all other actions require connecting to AWS
    else:
        builder = StudyBuilder(
            args["region"],
            args["workgroup"],
            args["profile"],
            args["schema_name"],
        )
        if args["verbose"]:
            builder.verbose = True
        print("Testing connection to athena...")
        builder.cursor.execute("SHOW DATABASES")

        if args["action"] == "clean":
            builder.clean_study(args["target"])

        else:
            study_dict = get_study_dict(args["study_dir"])
            if args["target"]:
                for target in args["target"]:
                    if target not in study_dict:
                        sys.exit(
                            f"{target} was not found in available studies: "
                            f"{list(study_dict.keys())}.\n\n"
                            "If you are trying to run a custom study, make sure "
                            "you include `-s path/to/study/dir` as an arugment."
                        )

            if args["action"] == "build":
                if "all" in args["target"]:
                    builder.clean_and_build_all(study_dict)
                else:
                    for target in args["target"]:
                        if args["table_builder"]:
                            builder.run_single_table_builder(
                                study_dict[target], args["table_builder"]
                            )
                        else:
                            builder.clean_and_build_study(study_dict[target])

            elif args["action"] == "export":
                if "all" in args["target"]:
                    builder.export_all(study_dict, args["data_path"])
                else:
                    for target in args["target"]:
                        builder.export_study(study_dict[target], args["data_path"])

        # returning the builder for ease of unit testing
        return builder


def main(cli_args=None):
    """Reads CLI input/environment variables and invokes library calls"""

    parser = get_parser()
    args = vars(parser.parse_args(cli_args))
    if args["version"]:
        print(__version__)
        sys.exit(0)
    if args["action"] is None:
        parser.print_usage()
        sys.exit(1)
    if args.get("target"):
        for target in args["target"]:
            if target == "all":
                args["target"] = ["all"]
                break

    arg_env_pairs = (
        ("profile", "CUMULUS_LIBRARY_PROFILE"),
        ("schema_name", "CUMULUS_LIBRARY_DATABASE"),
        ("workgroup", "CUMULUS_LIBRARY_WORKGROUP"),
        ("region", "CUMULUS_LIBRARY_REGION"),
        ("study_dir", "CUMULUS_LIBRARY_STUDY_DIR"),
        ("data_path", "CUMULUS_LIBRARY_DATA_PATH"),
        ("user", "CUMULUS_AGGREGATOR_USER"),
        ("id", "CUMULUS_AGGREGATOR_ID"),
        ("url", "CUMULUS_AGGREGATOR_URL"),
    )
    read_env_vars = []
    for pair in arg_env_pairs:
        if env_val := os.environ.get(pair[1]):
            args[pair[0]] = env_val
            read_env_vars.append([pair[1], env_val])

    if len(read_env_vars) > 0:
        table = Table(title="Values read from environment variables")
        table.add_column("Environment Variable", style="green")
        table.add_column("Value", style="cyan")
        for row in read_env_vars:
            if row[0] == "CUMULUS_AGGREGATOR_ID":
                table.add_row(row[0], "#########")
            else:
                table.add_row(row[0], row[1])
        console = Console()
        console.print(table)

    if args.get("study_dir"):
        posix_paths = []
        for path in args["study_dir"]:
            posix_paths.append(get_abs_posix_path(path))
        args["study_dir"] = posix_paths

    if args.get("data_path"):
        args["data_path"] = get_abs_posix_path(args["data_path"])
    return run_cli(args)


def main_cli():  # called by the generated wrapper scripts
    main()


if __name__ == "__main__":
    main()



""" Module for generating condition codeableConcept table"""

from cumulus_library.base_table_builder import BaseTableBuilder
from cumulus_library.template_sql.templates import (
    CodeableConceptConfig,
    get_codeable_concept_denormalize_query,
)


class ConditionCodableConceptBuilder(BaseTableBuilder):
    display_text = "Creating condition code table..."

    def prepare_queries(self, cursor: object, schema: str):
        """Constructs queries related to condition codeableConcept

        :param cursor: A database cursor object
        :param schema: the schema/db name, matching the cursor

        """
        config = CodeableConceptConfig(
            source_table="condition",
            source_id="id",
            cc_column={
                "name": "code",
                "is_array": False,
                "code_systems": [
                    "http://snomed.info/sct",
                    "http://hl7.org/fhir/sid/icd-10-cm",
                    "http://hl7.org/fhir/sid/icd-9-cm",
                ],
            },
            target_table="core__condition_codable_concepts",
        )
        self.queries.append(get_codeable_concept_denormalize_query(config))



from typing import List
from cumulus_library.schema import counts

STUDY_PREFIX = "core"


def table(tablename: str, duration=None) -> str:
    if duration:
        return f"{STUDY_PREFIX}__{tablename}_{duration}"
    else:
        return f"{STUDY_PREFIX}__{tablename}"


def count_patient():
    view_name = table("count_patient")
    from_table = table("patient")
    cols = ["age", "gender", "race_display", "ethnicity_display"]

    return counts.count_patient(view_name, from_table, cols)


def count_encounter(duration=None):
    view_name = table("count_encounter", duration)
    from_table = table("encounter")

    cols = [
        f"start_{duration}",
        "enc_class_display",
        "age_at_visit",
        "gender",
        "race_display",
        "ethnicity_display",
    ]

    return counts.count_encounter(view_name, from_table, cols)


def _count_encounter_type(view_name, cols, duration):
    """
    Encounter Type information is for every visit, and therefore this
    SQL should be precise in which fields to select (This is a BIG query).

    :param view_name: name of the view from "core__encounter_type"
    :param cols: from "core__encounter_type"
    :param duration: None or ''month', 'year'
    :return: SQL commands
    """
    view_name = table(view_name, duration)
    from_table = table("encounter_type")

    if duration:
        cols.append(f"start_{duration}")

    where = counts.where_clauses(min_subject=10)

    return counts.count_encounter(view_name, from_table, cols, where)


def count_encounter_type(duration=None):
    cols = [
        "enc_class_display",
        "enc_type_display",
        "enc_service_display",
        "enc_priority_display",
    ]
    return _count_encounter_type("count_encounter_type", cols, duration)


def count_encounter_enc_type(duration="month"):
    cols = ["enc_class_display", "enc_type_display"]
    return _count_encounter_type("count_encounter_enc_type", cols, duration)


def count_encounter_service(duration="month"):
    cols = ["enc_class_display", "enc_service_display"]
    return _count_encounter_type("count_encounter_service", cols, duration)


def count_encounter_priority(duration="month"):
    cols = ["enc_class_display", "enc_priority_display"]
    return _count_encounter_type("count_encounter_priority", cols, duration)


def concat_view_sql(create_view_list: List[str]) -> str:
    """
    :param create_view_list: SQL prepared statements
    :param filename: path to output file, default 'count.sql' in PWD
    """
    seperator = "-- ###########################################################"
    concat = list()

    for create_view in create_view_list:
        concat.append(seperator + "\n" + create_view + "\n")

    return "\n".join(concat)


def write_view_sql(view_list_sql: List[str], filename="count_core.sql") -> None:
    """
    :param view_list_sql: SQL prepared statements
    :param filename: path to output file, default 'count_core.sql' in PWD
    """
    sql_optimizer = concat_view_sql(view_list_sql)
    sql_optimizer = sql_optimizer.replace("ORDER BY cnt desc", "")
    sql_optimizer = sql_optimizer.replace("CREATE or replace VIEW", "CREATE TABLE")

    with open(filename, "w") as fout:
        fout.write(sql_optimizer)


if __name__ == "__main__":
    write_view_sql(
        [
            count_patient(),
            count_encounter("month"),
            count_encounter_type(),
            count_encounter_type("month"),
            count_encounter_enc_type("month"),
            count_encounter_service("month"),
            count_encounter_priority("month"),
        ]
    )



# pylint: disable=W,C,R
from enum import Enum

from fhirclient.models.fhirdate import FHIRDate


class FHIR(Enum):
    def __init__(self, url: str):
        self.url = url


class Structure(FHIR):
    """
    FHIR Conformance (partial conformance towards USCDI)
    """

    Patient = "http://hl7.org/fhir/us/core/StructureDefinition/us-core-patient"
    Encounter = "http://hl7.org/fhir/us/core/StructureDefinition/us-core-encounter"
    DocumentReference = (
        "http://hl7.org/fhir/us/core/StructureDefinition/us-core-documentreference"
    )
    Condition = "http://hl7.org/fhir/condition-definitions.html"
    ObservationLab = (
        "http://hl7.org/fhir/us/core/StructureDefinition/us-core-observation-lab"
    )
    ObservationValue = (
        "http://hl7.org/fhir/observation-definitions.html#Observation.value_x_"
    )
    VitalSign = "http://hl7.org/fhir/us/vitals/ImplementationGuide/hl7.fhir.us.vitals"


class Vocab(FHIR):
    """
    Terminologies (mapped to UMLS) referenced by Cumulus
    https://terminology.hl7.org/
    """

    ICD9 = "http://hl7.org/fhir/sid/icd-9-cm"
    ICD10 = "http://hl7.org/fhir/sid/icd-10-cm"
    LOINC = "http://loinc.org"
    SNOMED = "http://snomed.info/sct"
    UMLS = "http://www.nlm.nih.gov/research/umls/"
    RXNORM = "http://www.nlm.nih.gov/research/umls/rxnorm"
    CPT = "http://www.ama-assn.org/go/cpt"


class Datatypes(FHIR):
    # basic datatypes
    Boolean = "https://hl7.org/fhir/datatypes.html#boolean"
    Date = "https://www.hl7.org/fhir/datatypes.html#date"
    DateTime = "https://www.hl7.org/fhir/datatypes.html#dateTime"
    Str = "https://hl7.org/fhir/datatypes.html#string"
    Int = "http://hl7.org/fhir/datatypes.html#positiveInt"
    Decimal = "https://www.hl7.org/fhir/datatypes.html#decimal"
    Code = "https://www.hl7.org/fhir/datatypes.html#code"
    Coding = "https://www.hl7.org/fhir/datatypes.html#coding"
    Range = "https://hl7.org/fhir/datatypes.html#Range"
    Period = "https://www.hl7.org/fhir/datatypes.html#Period"
    Duration = "http://hl7.org/fhir/datatypes.html#Duration"
    Count = "http://hl7.org/fhir/search.html#count"


class Period:
    def __init__(self, start=None, end=None):
        """
        :param start: date
        :param end: date
        """
        self.system = Datatypes.Period
        self.start = None
        self.end = None

        if isinstance(start, str):
            self.start = FHIRDate(start)
        else:
            self.start = self.start

        if isinstance(end, str):
            self.end = FHIRDate(end)
        else:
            self.end = end

    def as_json(self):
        return {"start": str(self.start), "end": str(self.end), "system": self.system}


class Range:
    def __init__(self, low=None, high=None):
        self.system = Datatypes.Range
        self.low = low
        self.high = high

    def as_json(self):
        return {"low": str(self.low), "high": str(self.high), "system": self.system}


class Coding(Enum):
    def __init__(self, code=None, display=None, system=None):
        self.system = system
        self.code = code
        self.display = display if display else code

    def as_json(self):
        if self.system:
            return {"code": self.code, "display": self.display, "system": self.system}
        else:
            return {"code": self.code, "display": self.display}



""" Module for extracting US core extensions from patient records"""
from cumulus_library.base_table_builder import BaseTableBuilder
from cumulus_library.template_sql.templates import (
    get_extension_denormalize_query,
    ExtensionConfig,
)


class PatientExtensionBuilder(BaseTableBuilder):
    display_text = "Creating patient extension tables..."

    def prepare_queries(self, cursor: object, schema: str):
        """constructs queries related to patient extensions of interest

        :param cursor: A database cursor object
        :param schema: the schema/db name, matching the cursor
        """
        extension_types = [
            {
                "name": "race",
                "fhirpath": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-race",
            },
            {
                "name": "ethnicity",
                "fhirpath": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity",
            },
        ]

        for extension in extension_types:
            config = ExtensionConfig(
                "patient",
                "id",
                f"core__patient_ext_{extension['name']}",
                extension["name"],
                extension["fhirpath"],
                ["ombCategory", "detailed", "text"],
            )
            self.queries.append(get_extension_denormalize_query(config))



""" Handles pushing data to the aggregator"""
import sys

from pathlib import Path

import requests

from pandas import read_parquet
from rich.progress import Progress, TaskID

from cumulus_library.helper import get_progress_bar


def upload_data(
    progress: Progress,
    file_upload_progress: TaskID,
    file_path: Path,
    version: int,
    args: dict,
):
    """Fetches presigned url and uploads file to aggregator"""
    study = file_path.parts[-2]
    file_name = file_path.parts[-1]
    progress.update(file_upload_progress, description=f"Uploading {study}/{file_name}")
    data_package = file_name.split(".")[0]
    prefetch_res = requests.post(
        args["url"],
        json={
            "study": study,
            "data_package": data_package,
            "data_package_version": version,
            "filename": f"{args['user']}_{file_name}",
        },
        auth=(args["user"], args["id"]),
        timeout=60,
    )
    if args["preview"]:
        print("prefetch request")
        print("headers", prefetch_res.request.headers)
        print("body", prefetch_res.request.body, "\n")
        print("response")
        print(prefetch_res.json(), "\n")

    if prefetch_res.status_code != 200:
        print("Invalid user/site id")
        raise requests.RequestException(response=prefetch_res)
    res_body = prefetch_res.json()

    with open(file_path, "rb") as data_file:
        files = {"file": (file_name, data_file)}
        upload_req = requests.Request(
            "POST", res_body["url"], data=res_body["fields"], files=files
        ).prepare()
        if not args["preview"]:
            s = requests.Session()
            upload_res = s.send(upload_req, timeout=60)
            if upload_res.status_code != 204:
                print(f"Error uploading {study}/{file_name}")
                raise requests.RequestException(response=upload_res)
        else:
            print("upload_req")
            print("headers", upload_req.headers)
            print("body", upload_req.body, "\n")
    progress.update(file_upload_progress, advance=1)


def upload_files(args: dict):
    """Wrapper to prep files & console output"""
    if args["data_path"] is None:
        sys.exit(
            "No data directory provided - please provide a path to your"
            "study export folder."
        )
    file_paths = list(args["data_path"].glob("**/*.parquet"))
    if not args["user"] or not args["id"]:
        sys.exit("user/id not provided, please pass --user and --id")
    try:
        meta_version = next(
            filter(lambda x: str(x).endswith("__meta_version.parquet"), file_paths)
        )
        version = str(read_parquet(meta_version)["data_package_version"][0])
        file_paths.remove(meta_version)
    except StopIteration:
        version = "0"
    num_uploads = len(file_paths)
    with get_progress_bar() as progress:
        file_upload_progress = progress.add_task("Uploading", total=num_uploads)
        for file_path in file_paths:
            upload_data(progress, file_upload_progress, file_path, version, args)



import os
from typing import List

from fhirclient.models.coding import Coding

from cumulus_library.helper import load_json
from cumulus_library.template_sql.templates import get_create_view_query


def get_include_coding(valueset_json) -> List[Coding]:
    """
    Obtain a list of Coding "concepts" from a ValueSet.
    This method currently supports only "include" of "concept" defined fields.
    Not supported: recursive fetching of contained ValueSets, which requires UMLS API Key and Wget, etc.

    examples
    https://vsac.nlm.nih.gov/valueset/2.16.840.1.113762.1.4.1146.1629/expansion/Latest
    https://cts.nlm.nih.gov/fhir/res/ValueSet/2.16.840.1.113762.1.4.1146.1629?_format=json

    :param valueset_json: ValueSet file, expecially those provided by NLM/ONC/VSAC
    :return: list of codeable concepts (system, code, display) to include
    """
    valueset = load_json(valueset_json)
    parsed = []

    for include in valueset["compose"]["include"]:
        if "concept" in include.keys():
            for concept in include["concept"]:
                concept["system"] = include["system"]
                parsed.append(Coding(concept))
    return parsed


def create_view_sql(view_name: str, concept_list: List[Coding]) -> str:
    """
    :param view_name: like study__define_type
    :param concept_list: list of concepts to include in definition
    :return: sql statement to execute
    """
    if view_name is None:
        raise TypeError("No view name provided")
    content = []
    for concept in concept_list:
        content.append([concept.system, concept.code, concept.display])
    return get_create_view_query(
        view_name=view_name, dataset=content, view_cols=["system", "code", "display"]
    )


"""    header = f"create or replace view {view_name} as select * from (values"
    footer = ") AS t (system, code, display) ;"
    content = list()
    for concept in concept_list:
        content.append(f"('{concept.system}', '{concept.code}', '{concept.display}')")
    content = "\n,".join(content)
    return header + "\n" + content + "\n" + footer
"""


def write_view_sql(view_name: str, concept_list: List[Coding]) -> None:
    """Convenience wrapper for writing create_view_sql to disk"""
    with open(f"{view_name}.sql", "w") as fp:
        fp.write(create_view_sql(view_name, concept_list))



# pylint: disable=W,C,R
from cumulus_library.schema.valueset import DurationUnits
from cumulus_library.schema.columns import ColumnEnum as Column
from cumulus_library.errors import LibraryError

##################################################
# Google Style Guide
#
# https://google.github.io/styleguide/pyguide.html
# Design by composition composition for
#
# Query Grammar (in progress)
#
# Functions = support iterable lists/types
# Objects = typesafety only
##################################################


def drop(table_or_view) -> str:
    return f"DROP table if exists {table_or_view};"


def create_view(view_name) -> str:
    return f"CREATE or replace VIEW {view_name} AS "


def name_view(count_from_table: str, duration_col=None):
    if duration_col:
        return f"{count_from_table}_{as_duration(duration_col).name}"
    else:
        return f"{count_from_table}"


def as_duration(col: str) -> DurationUnits:
    if "_months" in col:
        return DurationUnits.months
    elif "_weeks" in col:
        return DurationUnits.weeks
    elif "_days" in col:
        return DurationUnits.days


def str_columns(cols) -> str:
    if isinstance(cols, str):
        return cols
    if isinstance(cols, Column):
        return cols.name
    if isinstance(cols, list):
        targets = [str_columns(c) for c in cols]
        return ", ".join(targets)


def where_clauses(clause=None, min_subject=10) -> str:
    if not clause:
        return f"cnt_subject >= {min_subject}"
    elif isinstance(clause, str):
        return clause
    elif isinstance(clause, list):
        return str_columns(clause) + ", cnt desc"
    else:
        raise LibraryError(f"where_sql() invalid clause {clause}")


def order_by_sql(order=None, cnt_desc=True) -> str:
    if not order:
        if cnt_desc:
            return "cnt desc;"
    elif isinstance(order, str):
        return order
    elif isinstance(order, list):
        if cnt_desc:
            return str_columns(order) + ", cnt desc;"
        else:
            return str_columns(order) + ";"


def count_patient(
    view_name: str, from_table: str, cols_object, where=None, order_by=None
) -> str:
    return count_query(
        view_name, from_table, cols_object, where, order_by, cnt_encounter=False
    )


def count_encounter(
    view_name: str, from_table: str, cols_object, where=None, order_by=None
) -> str:
    return count_query(
        view_name, from_table, cols_object, where, order_by, cnt_encounter=True
    )


def count_query(
    view_name: str,
    from_table: str,
    cols_object,
    where=None,
    order_by=None,
    cnt_encounter=None,
) -> str:
    cols = str_columns(cols_object)

    cnt_subject = "count(distinct subject_ref)   as cnt_subject"

    if cnt_encounter:
        cnt_encounter = ", count(distinct encounter_ref)   as cnt_encounter"
    else:
        cnt_encounter = False

    return f"""
    {create_view(view_name)}
    with powerset as
    (
        select
        {cnt_subject}
        {cnt_encounter if cnt_encounter else ''}
        , {cols}        
        FROM {from_table}
        group by CUBE
        ( {cols} )
    )
    select
          {'cnt_encounter ' if cnt_encounter else 'cnt_subject'} as cnt 
        , {cols}
    from powerset 
    WHERE {where_clauses(where)} 
    ORDER BY {order_by_sql(order_by)}
    """.strip()



""" Module for generating core medication table"""

from cumulus_library.base_table_builder import BaseTableBuilder
from cumulus_library.helper import get_progress_bar, query_console_output
from cumulus_library.template_sql.templates import (
    CodeableConceptConfig,
    get_core_medication_query,
    get_is_table_not_empty_query,
    get_column_datatype_query,
    get_ctas_empty_query,
)
from cumulus_library.template_sql.utils import is_codeable_concept_populated


class MedicationBuilder(BaseTableBuilder):
    display_text = "Creating core medication table..."

    def _check_data_in_fields(self, cursor, schema: str):
        """Validates whether either observed medication source is present"""

        data_types = {
            "inline": False,
            "by_contained_ref": False,
            "by_external_ref": False,
        }

        table = "medicationrequest"
        base_col = "medicationcodeableconcept"

        with get_progress_bar(transient=True) as progress:
            task = progress.add_task(
                "Detecting available medication sources...",
                total=7,
            )

            # inline medications from FHIR medication
            data_types["inline"] = is_codeable_concept_populated(
                schema, table, base_col, cursor
            )
            if data_types["inline"]:
                query = get_column_datatype_query(schema, table, base_col)
                cursor.execute(query)
                progress.advance(task)
                if "userselected" not in str(cursor.fetchone()[0]):
                    has_userselected = False
                else:
                    has_userselected = True
            else:
                has_userselected = False
            # Validating presence of FHIR medication requests
            query = get_is_table_not_empty_query(
                "medicationrequest", "medicationreference"
            )
            cursor.execute(query)
            progress.advance(task)
            if cursor.fetchone() is None:
                return data_types, has_userselected
            query = get_column_datatype_query(
                schema, "medicationrequest", "medicationreference"
            )
            cursor.execute(query)
            progress.advance(task)
            if "reference" not in cursor.fetchone()[0]:
                return data_types, has_userselected
            query = get_is_table_not_empty_query(
                "medicationrequest", "medicationreference.reference"
            )
            cursor.execute(query)
            progress.advance(task)
            if cursor.fetchone() is None:
                return data_types, has_userselected

            # checking med ref contents for our two linkage cases
            query = get_is_table_not_empty_query(
                "medicationrequest",
                "medicationreference.reference",
                conditions=["medicationreference.reference LIKE '#%'"],
            )
            cursor.execute(query)
            progress.advance(task)
            if cursor.fetchone() is not None:
                data_types["by_contained_ref"] = True
            query = get_is_table_not_empty_query(
                "medicationrequest",
                "medicationreference.reference",
                conditions=["medicationreference.reference LIKE 'Medication/%'"],
            )
            cursor.execute(query)
            progress.advance(task)
            if cursor.fetchone() is not None:
                data_types["by_external_ref"] = True

            return data_types, has_userselected

    def prepare_queries(self, cursor: object, schema: str) -> dict:
        """Constructs queries related to condition codeableConcept

        :param cursor: A database cursor object
        :param schema: the schema/db name, matching the cursor

        """
        medication_datasources, has_userselected = self._check_data_in_fields(
            cursor, schema
        )
        if (
            medication_datasources["inline"]
            or medication_datasources["by_contained_ref"]
            or medication_datasources["by_external_ref"]
        ):
            self.queries.append(
                get_core_medication_query(medication_datasources, has_userselected)
            )
        else:
            self.queries.append(
                get_ctas_empty_query(
                    schema,
                    "core__medication",
                    ["id", "resourcetype", "code", "ingredient"],
                )
            )



""" Module for directly loading ICD bsvs into athena tables """
import csv

from pathlib import Path

from cumulus_library.base_table_builder import BaseTableBuilder
from cumulus_library.helper import query_console_output, get_progress_bar
from cumulus_library.template_sql.templates import (
    get_ctas_query,
    get_insert_into_query,
)


class VocabIcdRunner(BaseTableBuilder):
    display_text = "Creating ICD vocab..."
    partition_size = 1200

    @staticmethod
    def clean_row(row, filename):
        """Removes non-SQL safe charatcers from the input row."""
        for i in range(len(row)):
            cell = str(row[i]).replace("'", "").replace(";", ",")
            row[i] = cell
        return row

    def prepare_queries(self, cursor: object, schema: str):
        """Creates queries for populating ICD vocab

        TODO: this would be a lot faster if we converted the bsv to parquet,
        uploaded that, and then created the table from an external datasource

        :param cursor: A database cursor object
        :param schema: the schema/db name, matching the cursor
        """

        table_name = "vocab__icd"
        icd_files = ["ICD10CM_2023AA", "ICD10PCS_2023AA", "ICD9CM_2023AA"]
        path = Path(__file__).parent

        headers = ["CUI", "TTY", "CODE", "SAB", "STR"]
        header_types = [f"{x} string" for x in headers]
        rows_processed = 0
        dataset = []
        created = False
        for filename in icd_files:
            with open(f"{path}/{filename}.bsv", "r") as file:
                # For the first row in the dataset, we want to coerce types from
                # varchar(len(item)) athena default to to an unrestricted varchar, so
                # we'll create a table with one row - this make the recast faster, and
                # lets us set the partition_size a little higher by limiting the
                # character bloat to keep queries under athena's limit of 262144.
                reader = csv.reader(file, delimiter="|")
                if not created:
                    row = self.clean_row(next(reader), filename)
                    self.queries.append(
                        get_ctas_query(
                            schema_name=schema,
                            table_name=table_name,
                            dataset=[row],
                            table_cols=headers,
                        )
                    )
                    created = True
                for row in reader:
                    row = self.clean_row(row, filename)
                    dataset.append(row)
                    rows_processed += 1
                    if rows_processed == self.partition_size:
                        self.queries.append(
                            get_insert_into_query(
                                table_name=table_name,
                                table_cols=headers,
                                dataset=dataset,
                            )
                        )
                        dataset = []
                        rows_processed = 0
                if rows_processed > 0:
                    self.queries.append(
                        get_insert_into_query(
                            table_name=table_name, table_cols=headers, dataset=dataset
                        )
                    )



"""utility functions related to populating jinja templates

Currently, these deal with various edge cases around complex nested objects in sql
and validating them so that the actual generated queries can be constructed more
simply. This includes, but is not limited, to:
    - Data missing entirely
    - Data present, but 'nullish' - with some structure but no value
    - Data with deep missing elements
    - Data which may or may not be in an array depending on context
"""
from rich.progress import Progress, Task

from cumulus_library.template_sql.templates import (
    get_column_datatype_query,
    get_is_table_not_empty_query,
)


def is_codeable_concept_populated(
    schema: str,
    table: str,
    base_col: str,
    cursor,
    coding_element="coding",
) -> bool:
    """Check db to see if codeableconcept data exists.

    Will execute several exploratory queries to see if the column in question
    can be queried naively.

    :param schema: The schema/database name
    :param table: The table to query against
    :param base_col: the place to start validation from.
        This can be a nested element, like column.object.code
    :param cursor: a PEP-249 compliant database cursor
    :param coding_element: the place inside the code element to look for coding info.
        default: 'coding' (and :hopefully: this is always right)
    :returns: a boolean indicating if valid data is present.
    """
    query = get_is_table_not_empty_query(table, base_col)
    cursor.execute(query)
    if cursor.fetchone() is None:
        return False

    query = get_column_datatype_query(schema, table, base_col)
    cursor.execute(query)
    if coding_element not in str(cursor.fetchone()[0]):
        return False

    query = get_is_table_not_empty_query(
        table,
        "t1.row1",
        [
            {
                "source_col": f"{base_col}.coding",
                "table_alias": "t1",
                "row_alias": "row1",
            }
        ],
    )
    cursor.execute(query)
    if cursor.fetchone() is None:
        return False
    return True


def is_codeable_concept_array_populated(
    schema: str,
    table: str,
    base_col: str,
    cursor,
    coding_element="coding",
) -> bool:
    """Check db to see if an array of codeableconcept data exists.

    Will execute several exploratory queries to see if the column in question
    can be queried naively. Will advance the associated progress's task by 3 steps.

    :param schema: The schema/database name
    :param table: The table to query against
    :param base_col: the place to start validation from.
        This can be a nested element, like column.object.code
    :param cursor: a PEP-249 compliant database cursor
    :param coding_element: the place inside the code element to look for coding info.
        default: 'coding' (and :hopefully: this is always right)
    :returns: a boolean indicating if valid data is present.
    """
    query = get_is_table_not_empty_query(table, base_col)
    cursor.execute(query)
    if cursor.fetchone() is None:
        return False

    query = get_column_datatype_query(schema, table, base_col)
    cursor.execute(query)
    if coding_element not in str(cursor.fetchone()[0]):
        return False

    query = get_is_table_not_empty_query(
        table,
        "t2.row2",
        [
            {
                "source_col": base_col,
                "table_alias": "t1",
                "row_alias": "row1",
            },
            {
                "source_col": "row1.coding",
                "table_alias": "t2",
                "row_alias": "row2",
            },
        ],
    )
    cursor.execute(query)
    if cursor.fetchone() is None:
        return False
    return True



"""Error types"""


class LibraryError(Exception):
    """
    Package level error
    """

    pass


class LibrarySchemaError(Exception):
    """
    Package level error
    """


class StudyManifestParsingError(Exception):
    """Basic error for StudyManifestParser"""



""" Module for generating encounter codeableConcept table"""

from cumulus_library.base_table_builder import BaseTableBuilder
from cumulus_library.helper import get_progress_bar, query_console_output
from cumulus_library.template_sql.templates import (
    CodeableConceptConfig,
    get_codeable_concept_denormalize_query,
    get_is_table_not_empty_query,
    get_ctas_empty_query,
)
from cumulus_library.template_sql.utils import (
    is_codeable_concept_array_populated,
    is_codeable_concept_populated,
)


class EncounterCodingBuilder(BaseTableBuilder):
    display_text = "Creating encounter codeableConcept tables..."

    def _check_data_in_fields(self, code_sources: list[dict], schema, cursor) -> dict:
        """checks if CodeableConcept fields actually have data available

        CodeableConcept fields are mostly optional in the FHIR spec, and may be arrays
        or single objects. Additionally, the null representation can be inconsistent,
        depending on how the data is provided from an EHR and how the ETL manages
        schema inference (wide, but not deep). We :could: try to find the data and
        just catch an error, but that would mask configuration errors/unexpected
        data patterns. So, instead, we are doing the following fussy operation:

        For each column we want to check for data:
        - Check to see if there is any data in a codeableConcept field
        - Check to see if the codeableConcept field contains a coding element
        - Check if that coding element contains non-null data

        The way we do this is slightly different depending on if the field is an
        array or not (generally requiring one extra level of unnesting).

        """
        # TODO: consider moving to a utility library if we have another case like
        # this one - it would probably involve splitting this into two paths, one
        # for array encodings and one for non-array encodings.
        # See builder_core_medication for a non-array example

        with get_progress_bar(transient=True) as progress:
            task = progress.add_task(
                "Detecting available encounter codaebleConcepts...",
                # Each column in code_sources requires at most 3 queries to
                # detect valid data is in the DB
                total=len(code_sources),
            )
            for code_source in code_sources:
                if code_source["is_array"]:
                    code_source["has_data"] = is_codeable_concept_array_populated(
                        schema, "encounter", code_source["name"], cursor
                    )
                else:
                    code_source["has_data"] = is_codeable_concept_populated(
                        schema, "encounter", code_source["name"], cursor
                    )
                progress.advance(task)
        return code_sources

    def prepare_queries(self, cursor: object, schema: str):
        """Constructs queries related to condition codeableConcept

        :param cursor: A database cursor object
        :param schema: the schema/db name, matching the cursor

        """

        code_sources = [
            {
                "name": "type",
                "is_array": True,
                "code_systems": [
                    "http://terminology.hl7.org/CodeSystem/encounter-type",
                    "http://terminology.hl7.org/CodeSystem/v2-0004",
                    "urn:oid:2.16.840.1.113883.4.642.3.248",
                    "http://snomed.info/sct",
                ],
                "has_data": False,
            },
            {
                "name": "servicetype",
                "is_array": False,
                "code_systems": [
                    "http://terminology.hl7.org/CodeSystem/service-type",
                    "urn:oid:2.16.840.1.113883.4.642.3.518",
                    "http://snomed.info/sct",
                ],
                "has_data": False,
            },
            {
                "name": "priority",
                "is_array": False,
                "code_systems": [
                    "http://terminology.hl7.org/CodeSystem/v3-ActPriority",
                    "http://snomed.info/sct",
                ],
                "has_data": False,
            },
        ]
        code_sources = self._check_data_in_fields(code_sources, schema, cursor)
        for code_source in code_sources:
            if code_source["has_data"]:
                config = CodeableConceptConfig(
                    source_table="encounter",
                    source_id="id",
                    cc_column=code_source,
                    target_table=f"core__encounter_dn_{code_source['name']}",
                )
                self.queries.append(get_codeable_concept_denormalize_query(config))
            else:
                self.queries.append(
                    get_ctas_empty_query(
                        schema_name=schema,
                        table_name=f"core__encounter_dn_{code_source['name']}",
                        table_cols=["id", "code", "code_system", "display"],
                    )
                )



# pylint: disable=W,C,R
from enum import Enum
from cumulus_library.schema.typesystem import Coding, Vocab

################################################################################
# FHIR ValueSets
################################################################################


class ValueSet(Enum):
    Gender = "http://hl7.org/fhir/ValueSet/administrative-gender"
    Race = "http://hl7.org/fhir/us/core/StructureDefinition/us-core-race"
    Ethnicity = "http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity"
    DurationUnits = "http://hl7.org/fhir/valueset-duration-units.html"
    Units = "http://hl7.org/fhir/ValueSet/ucum-units"
    PatientClass = "http://terminology.hl7.org/CodeSystem/v2-0004"
    EncounterCode = "http://terminology.hl7.org/ValueSet/v3-ActEncounterCode"
    EncounterPriority = "http://terminology.hl7.org/CodeSystem/v3-ActPriority"
    EncounterStatus = "http://hl7.org/fhir/ValueSet/encounter-status"
    EncounterType = "http://hl7.org/fhir/ValueSet/encounter-type"
    EncounterReason = "http://hl7.org/fhir/ValueSet/encounter-reason"
    EncounterLocationStatus = "http://hl7.org/fhir/ValueSet/encounter-location-status"
    ConditionCode = "http://hl7.org/fhir/ValueSet/condition-code"
    ConditionCategory = "http://hl7.org/fhir/ValueSet/condition-category"
    DocumentType = "http://hl7.org/fhir/ValueSet/c80-doc-typecodes"
    ObservationCode = "http://hl7.org/fhir/ValueSet/observation-codes"
    ObservationCategory = "http://hl7.org/fhir/ValueSet/observation-category"
    ObservationInterpretation = (
        "http://hl7.org/fhir/ValueSet/observation-interpretation"
    )

    def __init__(self, url: str):
        self.url = url


class DurationUnits(Coding):
    milliseconds = ("ms", "milliseconds")
    seconds = ("s", "seconds")
    minutes = ("min", "minutes")
    hours = ("h", "hours")
    days = ("d", "days")
    weeks = ("wk", "weeks")
    months = ("mo", "months")
    years = ("a", "years")

    def __init__(self, code, display):
        super().__init__(code, display, ValueSet.DurationUnits.url)


class Gender(Coding):
    """
    Biological Sex of patient, not "gender identity"
    http://hl7.org/fhir/valueset-administrative-gender.html
    """

    male = ("male", "Male")
    female = ("female", "Female")
    trans_male = ("other", "Other")
    unknown = ("unknown", "Unknown")

    def __init__(self, code=None, display=None):
        super().__init__(code, display, ValueSet.Gender.url)


class Race(Coding):
    """
    Race coding has 5 "root" levels, called the R5 shown below.
    http://hl7.org/fhir/r4/v3/Race/cs.html
    """

    asian = ("2028-9", "Asian")
    black = ("2054-5", "Black or African American")
    white = ("2106-3", "White")
    native = ("1002-5", "American Indian or Alaska Native")
    islander = ("2076-8", "Native Hawaiian or Other Pacific Islander")

    def __init__(self, code, display):
        super().__init__(code, display, ValueSet.Race.url)


class Ethnicity(Coding):
    """
    RWD usually has only this binary YES/NO hispanic or latino feature populated.
    For a complete list of codes, see Ethnicity.system.
    """

    hispanic_or_latino = ("2135-2", "Hispanic or Latino	Hispanic or Latino")
    not_hispanic_or_latino = ("2186-5", "Not Hispanic or Latino")

    def __init__(self, code, display):
        super().__init__(code, display, ValueSet.Ethnicity.url)


################################################################################
# Encounter
################################################################################


class EncounterCode(Coding):
    AMB = ("AMB", "ambulatory")
    EMER = ("EMER", "emergency")
    FLD = ("FLD", "field")
    HH = ("HH", "home health")
    IMP = ("IMP", "inpatient encounter")
    ACUTE = ("ACUTE", "inpatient acute")
    NONAC = ("NONAC", "inpatient non-acute")
    OBSENC = ("OBSENC", "observation encounter")
    PRENC = ("PRENC", "pre-admission")
    SS = ("SS", "short stay")
    VR = ("VR", "virtual")

    def __init__(self, code, display=None):
        super().__init__(code, display, ValueSet.EncounterCode.value)


################################################################################
# Observation
################################################################################


class ObservationInterpretationDetection(Coding):
    positive = ("POS", "Positive")
    negative = ("NEG", "Negative")
    indeterminate = ("IND", "Indeterminate")
    equivocal = ("E", "Equivocal")
    detected = ("DET", "Detected")
    not_detected = ("ND", "Not detected")

    def __init__(self, code, display):
        """
        Cumulus Library Note: PCR testing should use "POS", "NEG", and "IND" codes.
        http://hl7.org/fhir/R4/v3/ObservationInterpretation/cs.html#v3-ObservationInterpretation-ObservationInterpretationDetection

        Interpretations of the presence or absence of a component / analyte or organism in a test or of a sign in a clinical observation.
        In keeping with laboratory data processing practice, these concepts provide a categorical interpretation of the "meaning" of the quantitative value for the same observation.

        POS =
        A presence finding of the specified component / analyte, organism or clinical sign based on the established threshold of the performed test or procedure.

        NEG =
        An absence finding of the specified component / analyte, organism or clinical sign based on the established threshold of the performed test or procedure.

        IND =
        The specified component / analyte, organism or clinical sign could neither be declared positive / negative nor detected / not detected by the performed test or procedure.
        Usage Note: For example, if the specimen was degraded, poorly processed, or was missing the required anatomic structures, then "indeterminate" (i.e. "cannot be determined") is the appropriate response, not "equivocal".
        """
        super().__init__(code, display, ValueSet.ObservationInterpretation.url)



# pylint: disable=W,C,R
from enum import Enum, EnumMeta
from cumulus_library.schema.typesystem import Datatypes, Coding, Vocab
from cumulus_library.schema.valueset import Gender, Race, Ethnicity
from cumulus_library.schema.valueset import DurationUnits
from cumulus_library.schema.valueset import EncounterCode
from cumulus_library.schema.valueset import ObservationInterpretationDetection


class ColumnEnum(Enum):
    # Counts
    cnt = Datatypes.Int, "Count"
    cnt_subject = Datatypes.Int, "Count Patients"
    cnt_encounter = Datatypes.Int, "Count Encounters"
    cnt_document = Datatypes.Int, "Count Documents"
    cnt_observation = Datatypes.Int, "Count Observations"
    cnt_lab = Datatypes.Int, "Count Lab Tests"

    # DataType Generics
    code_display = Datatypes.Str, "Code Display"
    period = Datatypes.Period, "Date Period"

    # Patient
    gender = Datatypes.Coding, "Biological sex at birth", Gender
    race = Datatypes.Coding, "Patient reported race", Race
    race_code = Datatypes.Coding, "Patient reported race", Race
    race_display = Datatypes.Str, "Patient reported race", Race
    ethnicity = Datatypes.Coding, "Patient reported ethnicity", Ethnicity
    age = Datatypes.Int, "Age in years calculated since DOB"
    postalcode3 = Datatypes.Int, "Patient 3 digit zip"

    # Encounter
    age_at_visit = Datatypes.Int, "Patient Age at Encounter"
    age_group = Datatypes.Str, "Patient Age Group at Encounter"
    enc_class = Datatypes.Coding, "Encounter Code (Healthcare Setting)", EncounterCode
    enc_class_code = (
        Datatypes.Coding,
        "Encounter Code (Healthcare Setting)",
        EncounterCode,
    )
    start_date = Datatypes.DateTime, "Day patient encounter started", DurationUnits.days
    start_week = Datatypes.Date, "Week patient encounter started", DurationUnits.weeks
    start_month = (
        Datatypes.Date,
        "Month patient encounter started",
        DurationUnits.months,
    )
    start_year = Datatypes.Date, "Year patient encounter started", DurationUnits.years
    end_date = Datatypes.DateTime, "Day patient encounter ended", DurationUnits.days
    enc_los_days = Datatypes.Int, "LOS Length Of Stay (days)", DurationUnits.days
    enc_los_weeks = Datatypes.Int, "LOS Length Of Stay (weeks)", DurationUnits.weeks

    # Condition
    cond_code = Datatypes.Coding, "Condition code"
    cond_code_display = Datatypes.Str, "Condition code"
    cond_icd10 = Vocab.ICD10, "Condition code ICD10"
    cond_snomed = Vocab.SNOMED, "Condition code SNOMED"
    cond_date = Datatypes.DateTime, "Day condition recorded", DurationUnits.days
    cond_week = Datatypes.Date, "Week condition recorded", DurationUnits.weeks
    cond_month = Datatypes.Date, "Month condition recorded", DurationUnits.months
    cond_year = Datatypes.Date, "Year condition recorded", DurationUnits.years

    # DocumentReference
    author_date = Datatypes.DateTime, "Day document was authored", DurationUnits.days
    author_week = Datatypes.Date, "Week document was authored", DurationUnits.weeks
    author_month = Datatypes.Date, "Month document was authored", DurationUnits.months
    author_year = Datatypes.Date, "Year document was authored", DurationUnits.years
    doc_type = Vocab.LOINC, "Type of Document"
    doc_type_code = Datatypes.Coding, "Type of Document (code)"
    doc_type_display = Datatypes.Str, "Type of Document (display)"
    doc_class = Vocab.LOINC, "Class of DocumentNote"
    ed_note = Datatypes.Boolean, "ED Note was documented for encounter (true/false)"

    # Observation Lab
    lab_code = Datatypes.Coding, "Laboratory Code"
    lab_code_display = Datatypes.Str, "Laboratory Code Display"
    lab_result = (
        Datatypes.Coding,
        "Laboratory result interpretation",
        ObservationInterpretationDetection,
    )
    lab_result_display = Datatypes.Str, "Laboratory result"
    lab_date = Datatypes.DateTime, "Day of lab result", DurationUnits.days
    lab_week = Datatypes.Date, "Week of lab result", DurationUnits.weeks
    lab_month = Datatypes.Date, "Month of lab result", DurationUnits.months
    lab_year = Datatypes.Date, "Year of lab result", DurationUnits.years
    analyte = Datatypes.Str, "Analyte/component"

    # Observation Lab PCR
    pcr_code = Datatypes.Coding, "PCR test code"
    pcr_result = (
        Datatypes.Coding,
        "PCR result interpretation",
        ObservationInterpretationDetection,
    )
    pcr_result_display = (
        Datatypes.Str,
        "PCR result interpretation",
        ObservationInterpretationDetection,
    )
    pcr_date = Datatypes.Date, "Day of PCR result", DurationUnits.days
    pcr_week = Datatypes.Date, "Week of PCR result", DurationUnits.weeks
    pcr_month = Datatypes.Date, "Month of PCR result", DurationUnits.months
    pcr_year = Datatypes.Date, "Year of PCR result", DurationUnits.years



""" Collection of small commonly used utility functions """
import os
import json
from typing import List
from rich import progress


def filepath(filename: str) -> str:
    return os.path.join(os.path.dirname(__file__), filename)


def load_text(path: str) -> str:
    with open(path, "r", encoding="UTF-8") as fp:
        return fp.read()


def load_json(path: str) -> dict:
    with open(path, "r", encoding="UTF-8") as fp:
        return json.load(fp)


def parse_sql(sql_text: str) -> List[str]:
    commands = []

    for statement in sql_text.split(";"):
        parsed = []
        for line in statement.splitlines():
            if not line.strip().startswith("--"):
                parsed.append(line)
        commands.append("\n".join(parsed))
    return filter_strip(commands)


def filter_strip(commands) -> List[str]:
    return list(filter(None, [c.strip() for c in commands]))


def list_coding(code_display: dict, system=None) -> List[dict]:
    as_list = []
    for code, display in code_display.items():
        if system:
            item = {"code": code, "display": display, "system": system}
        else:
            item = {"code": code, "display": display}
        as_list.append(item)
    return as_list


def query_console_output(
    verbose: bool, query: str, progress_bar: progress.Progress, task: progress.Task
):
    """Convenience function for determining output type"""
    if verbose:
        print()
        print(query)
    else:
        progress_bar.advance(task)


def get_progress_bar(**kwargs) -> progress.Progress:
    # The default columns don't change to elapsed time when finished.
    return progress.Progress(
        progress.TextColumn("[progress.description]{task.description}"),
        progress.BarColumn(),
        progress.TaskProgressColumn(),
        progress.TimeRemainingColumn(elapsed_when_finished=True),
        **kwargs,
    )



""" abstract base for python-based study executors """
import re

from abc import ABC, abstractmethod
from pathlib import Path
from typing import final

from cumulus_library.helper import get_progress_bar, query_console_output


class BaseTableBuilder(ABC):
    """Generic base class for python table builders.

    To use a table builder, extend this class exactly once in a new module.
    See ./studies/vocab or ./studies/core for example usage.
    """

    display_text = "Building custom tables..."

    def __init__(self):
        self.queries = []

    @abstractmethod
    def prepare_queries(self, cursor: object, schema: str):
        """Main entrypoint for python table builders.

        When completed, prepare_queries should populate self.queries with sql
        statements to execute. This array will the be read by run queries.

        :param cursor: A PEP-249 compatible cursor
        :param schema: A schema name
        :param verbose: toggle for verbose output mode
        """
        raise NotImplementedError

    @final
    def execute_queries(
        self, cursor: object, schema: str, verbose: bool, drop_table: bool = False
    ):
        """Executes queries set up by a prepare_queries call

        :param cursor: A PEP-249 compatible cursor
        :param schema: A schema name
        :param verbose: toggle for verbose output mode
        :param drop_table: drops any tables found in prepared_queries results
        """
        self.prepare_queries(cursor, schema)
        if drop_table:
            table_names = []
            for query in self.queries:
                # Get the first non-whitespace word after create table
                table_name = re.search(
                    '(?i)(?<=create table )(([a-zA-Z0-9_".-]+))', query
                )  # [0]
                if table_name:
                    table_name = table_name[0]
                    # if it contains a schema, remove it (usually it won't, but some CTAS
                    # forms may)
                    if "." in table_name[0]:
                        table_name = table_name.split(".")[1].replace('"', "")
                    table_names.append(table_name)
            for table_name in table_names:
                cursor.execute(f"DROP TABLE IF EXISTS {table_name}")
        with get_progress_bar(disable=verbose) as progress:
            task = progress.add_task(
                self.display_text,
                total=len(self.queries),
                visible=not verbose,
            )
            for query in self.queries:
                cursor.execute(query)
                query_console_output(verbose, self.queries, progress, task)

    def write_queries(self, filename: str = "output.sql"):
        with open(filename, "w") as file:
            for query in self.queries:
                file.write(query)
                file.write("\n")



""" Collection of jinja template getters for common SQL queries """
from enum import Enum
from pathlib import Path
from typing import Dict, List

from jinja2 import Template


class TableView(Enum):
    TABLE = "TABLE"
    VIEW = "VIEW"


def get_drop_view_table(name: str, view_or_table: str) -> str:
    """Generates a drop table if exists query"""
    if view_or_table in [e.value for e in TableView]:
        path = Path(__file__).parent
        with open(f"{path}/drop_view_table.sql.jinja") as drop_view_table:
            return Template(drop_view_table.read()).render(
                view_or_table_name=name, view_or_table=view_or_table
            )


def get_show_tables(schema_name: str, prefix: str) -> str:
    """Generates a show tables query, filtered by prefix

    The intended use case for this function is to get a list of manifest study
    tables, so that they can be individually dropped during a clean call.

    :param schema_name: The athena schema to query
    :param table_name: The prefix to filter by. Jinja template auto adds '__'.
    """
    path = Path(__file__).parent
    with open(f"{path}/show_tables.sql.jinja") as show_tables:
        return Template(show_tables.read()).render(
            schema_name=schema_name, prefix=prefix
        )


def get_show_views(schema_name: str, prefix: str) -> str:
    """Generates a show vies query, filtered by prefix

    The intended use case for this function is to get a list of manifest study
    views, so that they can be individually dropped during a clean call.

    :param schema_name: The athena schema to query
    :param table_name: The prefix to filter by. Jinja template auto adds '__'.
    """
    path = Path(__file__).parent
    with open(f"{path}/show_views.sql.jinja") as show_tables:
        return Template(show_tables.read()).render(
            schema_name=schema_name, prefix=prefix
        )


def get_ctas_query(
    schema_name: str, table_name: str, dataset: List[List[str]], table_cols: List[str]
) -> str:
    """Generates a create table as query for inserting static data into athena

    Note that unlike other queries, the nature of the CTAS implementation in athena
    requires a schema name. This schema name should match the schema of your cursor,
    or the other queries in this template will not function correctly. All columns
    will be specified as varchar type.

    :param schema_name: The athena schema to create the table in
    :param table_name: The name of the athena table to create
    :param dataset: Array of data arrays to insert, i.e. [['1','3'],['2','4']]
    :param table_cols: Comma deleniated column names, i.e. ['first,second']
    """
    path = Path(__file__).parent
    with open(f"{path}/ctas.sql.jinja") as ctas:
        return Template(ctas.read()).render(
            schema_name=schema_name,
            table_name=table_name,
            dataset=dataset,
            table_cols=table_cols,
        )


def get_ctas_empty_query(
    schema_name: str, table_name: str, table_cols: List[str]
) -> str:
    """Generates a create table as query for initializing an empty table

    Note that unlike other queries, the nature of the CTAS implementation in athena
    requires a schema name. This schema name should match the schema of your cursor,
    or the other queries in this template will not function correctly. All columns
    will be specified as varchar type.

    :param schema_name: The athena schema to create the table in
    :param table_name: The name of the athena table to create
    :param table_cols: Comma deleniated column names, i.e. ['first,second']
    """
    path = Path(__file__).parent
    with open(f"{path}/ctas_empty.sql.jinja") as ctas_empty:
        return Template(ctas_empty.read()).render(
            schema_name=schema_name,
            table_name=table_name,
            table_cols=table_cols,
        )


def get_create_view_query(
    view_name: str, dataset: List[List[str]], view_cols: List[str]
) -> str:
    """Generates a create view as query for inserting static data into athena

    :param view_name: The name of the athena table to create
    :param dataset: Array of data arrays to insert, i.e. [['1','3'],['2','4']]
    :param table_cols: Comma deleniated column names, i.e. ['first,second']
    """
    path = Path(__file__).parent
    with open(f"{path}/create_view_as.sql.jinja") as cvas:
        return Template(cvas.read()).render(
            view_name=view_name,
            dataset=dataset,
            view_cols=view_cols,
        )


def get_insert_into_query(
    table_name: str, table_cols: List[str], dataset: List[List[str]]
) -> str:
    """Generates an insert query for adding data to an existing athena table

    :param schema_name: The athena query to create the table in
    :param table_name: The name of the athena table to create
    :param table_cols: Comma deleniated column names, i.e. ['first','second']
    :param dataset: Array of data arrays to insert, i.e. [['1','3'],['2','4']]
    """
    path = Path(__file__).parent
    with open(f"{path}/insert_into.sql.jinja") as insert_into:
        return Template(insert_into.read()).render(
            table_name=table_name, table_cols=table_cols, dataset=dataset
        )


class ExtensionConfig(object):
    """convenience class for holding parameters for generating extension tables.

    :param source_table: the table to extract extensions from
    :param source_id: the id column to treat as a foreign key
    :param target_table: the name of the table to create
    :param target_col_prefix: the string to prepend code/display column names with
    :param fhir_extension: the URL of the FHIR resource to select
    :param code_systems: a list of codes, in preference order, to use to select data
    """

    def __init__(
        self,
        source_table: str,
        source_id: str,
        target_table: str,
        target_col_prefix: str,
        fhir_extension: str,
        ext_systems: List[str],
    ):
        self.source_table = source_table
        self.source_id = source_id
        self.target_table = target_table
        self.target_col_prefix = target_col_prefix
        self.fhir_extension = fhir_extension
        self.ext_systems = ext_systems


def get_extension_denormalize_query(config: ExtensionConfig) -> str:
    """extracts target extension from a table into a denormalized table

    This function is targeted at a complex extension element that is at the root
    of a FHIR resource - as an example, see the 5 codes at the root node of
    http://hl7.org/fhir/us/core/STU6/StructureDefinition-us-core-patient.html.
    The template will create a new table with the extension data, in arrays,
    mapped 1-1 to the table id. You can specify multiple systems
    in the ExtensionConfig passed to this function. For each patient, we'll
    take the data from the first extension coding system we find for each patient.

    :param config: An instance of ExtensionConfig.
    """
    path = Path(__file__).parent
    with open(f"{path}/extension_denormalize.sql.jinja") as extension_denormalize:
        return Template(extension_denormalize.read()).render(
            source_table=config.source_table,
            source_id=config.source_id,
            target_table=config.target_table,
            target_col_prefix=config.target_col_prefix,
            fhir_extension=config.fhir_extension,
            ext_systems=config.ext_systems,
        )


class CodeableConceptConfig:
    """Convenience class for holding parameters for generating codableconcept tables.

    :param source_table: the table to extract extensions from
    :param source_id: the id field to use in the new table
    :param cc_columns: the column containing the codeableConcept you want to extract.
        Format:
            {'name':[column],
            'is_array': [boolean],
            'code_systems':[List of code system strings, in priority order]}
        is_array relates to the FHIR spec - if the field is specified
        as 0...*, set this to be true.
    :param target_table: the name of the table to create
    :param code_systems: a list of systems, in preference order, for selecting data
    """

    def __init__(
        self, source_table: str, source_id: str, cc_column: dict, target_table: str
    ):
        self.source_table = source_table
        self.source_id = source_id
        self.cc_column = cc_column
        self.target_table = target_table


def get_codeable_concept_denormalize_query(config: CodeableConceptConfig) -> str:
    """extracts codeable concepts from a specified table.

    This function is targeted at arbitrary codeableConcept elements - see
    http://hl7.org/fhir/datatypes-definitions.html#CodeableConcept for more info.
    This may be or may not be an array field depending on the context of use -
    check the specification of the specific resource you're interested in.
    See the CodeableConceptConfig for details on how to handle array vs non-
    array use cases.

    :param config: a CodableConeptConfig
    """
    path = Path(__file__).parent
    with open(f"{path}/codeable_concept_denormalize.sql.jinja") as codable_concept:
        return Template(codable_concept.read()).render(
            source_table=config.source_table,
            source_id=config.source_id,
            cc_column=config.cc_column,
            target_table=config.target_table,
        )


def get_is_table_not_empty_query(
    source_table: str, field: str, unnests: list[dict] = [], conditions: list[str] = []
):
    path = Path(__file__).parent
    with open(f"{path}/is_table_not_empty.sql.jinja") as is_table_not_empty:
        return Template(is_table_not_empty.read()).render(
            source_table=source_table,
            field=field,
            unnests=unnests,
            conditions=conditions,
        )


def get_core_medication_query(
    medication_datasources: dict, has_userselected: bool = False
):
    path = Path(__file__).parent
    with open(f"{path}/core_medication.sql.jinja") as core_medication:
        return Template(core_medication.read()).render(
            medication_datasources=medication_datasources,
            has_userselected=has_userselected,
        )


def get_column_datatype_query(schema_name: str, table_name: str, column_name: str):
    path = Path(__file__).parent
    with open(f"{path}/column_datatype.sql.jinja") as column_datatype:
        return Template(column_datatype.read()).render(
            schema_name=schema_name,
            table_name=table_name,
            column_name=column_name,
        )
